/** @file fixnormal.c
 *  Correct inverted y components in normal maps. 
 *  This file contains code that will load a normal map contained within a
 *  Targa file, flip the y component of the normals (invert the green channel)
 *  and save the normal map back to a file. If an output name is specified, the
 *  result is written to that file, otherwise the source file is overwritten.
 *
 *  \author Chris <chris@starforge.co.uk>
 *  \copy 2008, Chris Page &lt;chris@starforge.co.uk&gt;
 *  \version 1.0
 *  \date 2008-10-30
 */
/*  
 *  Copyright (C) 2008 Chris Page <chris@starforge.co.uk>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be entertaining,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.  A copy of the
 *  General Public License is included in the file COPYING.
 */

#include <getopt.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "targa.h"

static const char *optstring = "o:vh?"; //!< getopt_long options string

/** Options structure for getopt_long. */
static const struct option longopts[] = {
    { "output" , required_argument, NULL, 'o' },
    { "verbose", no_argument      , NULL, 'v' },
    { "help"   , no_argument      , NULL, 'h' },
    { NULL     , no_argument      , NULL,   0 }
};

/** Invert the green channel of an image. Normalmaps store the normal vector
 *  x, y, and z in the r, g, and b channels of an image respectively. Some
 *  normal map generation tools generate normals with inverted y components,
 *  this function corrects the direction of the y component.
 *
 * \param image   The image containing the normalmap to fix.
 * \param verbose If true, additional status information is printed to stdout.
 */
static void invert_y(Image *image, int verbose)
{
    int pixel;

    // Process each pixel in the image
    for(pixel = 0; pixel < (image -> width * image -> height); ++pixel) {
        // Invert the green channel (flip the normal y)
        image -> data[pixel].g = 128 + (-1 * (image -> data[pixel].g - 128));
    }
}


/** Display the program usage and then exit.
 */
static void display_usage(void)
{
    puts("usage: fixnormal [options] FILE\n\n"  \
         "Options:\n" \
         "  -o, --output FILE    Write the fixed normal map to FILE (if not specified,\n" \
         "                       the original file is overwritten).\n"  \
         "  -v, --verbose        Display processing information.\n" \
         "  -h, --help           Display this help and exit.\n");
    exit(1);
}


int main(int argc, char *argv[])
{
    char *output = NULL;
    int verbose = 0, opt = 0, longindex = 0;

    // Process the command line with getopt_long
    while((opt = getopt_long(argc, argv, optstring, longopts, &longindex)) != -1) {
        switch(opt) {
            case 'o': output = optarg;
                break;
            case 'v': verbose = 1;
                break;
            case 'h': /* fall-through is intentional */
            case '?': display_usage();
                break;
            default: /* This should not happen */
                break;
        }
    }

    // If we have no file to process, fail with usage instructions
    if((argc - optind) == 0) {
        fprintf(stderr, "No targa file specified.\n");
        display_usage();
    }
    
    // If no output name has been specified, use the input name
    if(!output) {
        if(verbose) printf("No output file specified, the source file will be overwritten.\n"); 
        output = argv[optind];
    }
    
    // Load the file
    Image *process = loadTGA(argv[optind], verbose);

    if(process) {
        // Invert the normals
        invert_y(process, verbose);

        // And save..
        saveTGA(output, process, verbose);

        // Clean up
        free_image(process);
    }

    return 0;
}
