/** @file targa.c
 *  Targa file loading and saving functions.
 *  This file contains functions that allow image data to be loaded from, or saved to,
 *  Targa files. The range of supported Targa formats is limited in this code - it will
 *  only load 12, 24 or 32 bit type 2 (uncompressed RGB) TGA files, and no support is
 *  provided for compressed or colourmapped TGA formats.
 *
 *  \author Chris <chris@starforge.co.uk>
 *  \copy 2008, Chris Page &lt;chris@starforge.co.uk&gt;
 *  \version 1.0
 *  \date 2008-10-30
 */
/*  
 *  Copyright (C) 2008 Chris Page <chris@starforge.co.uk>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be entertaining,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.  A copy of the
 *  General Public License is included in the file COPYING.
 */

#include <errno.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include "targa.h"

/** Flip the specified image vertically. This is needed for images whose
 *  origin is in the top left rather than the bottom left used in 
 *  Truevision Targa files.
 *
 * \param Image the image to flip vertically.
 */
static void flip_vertically(Image *image)
{
    uint8_t *buffer;
    int linelen = sizeof(Pixel) * image -> width;
    // First allocate enough space for a copy of one line of the image
    if(!(buffer = malloc(linelen))) {
        fprintf(stderr, "Unable to allocate %d pixels for flip buffer. Image not flipped!\n", sizeof(Pixel) * image -> width);
        return;
    }

    // Now we need to flip the image vertically 
    int ytop = 0;
    int ybot = (image -> height - 1);
    while(ytop < ybot) {
         // Copy the top line into the buffer 
        memcpy(buffer, &image -> data[ytop * image -> width], linelen);
        // Copy the bottom to the top
        memcpy(&image -> data[ytop * image -> width], &image -> data[ybot * image -> width], linelen);
        // copy the buffer into the bottom
        memcpy(&image -> data[ybot * image -> width], buffer, linelen);

        // Move the top down and bottom up
        ++ytop; --ybot;
    }

    // Release the buffer
    free(buffer);
}


/** Load the header into memory, and place the parts we are interested
 *  in into the provided Image structure. If this is successful then 
 *  the width, height, and bpp fields of the Image structure are filled
 *  in with the corresponding values from the header, and the file
 *  pointer will be placed at the start of the image data in the file.
 *  If this fails (either a read error, or the image is not a type 2
 *  TGA) then this returns false, and prints the error code accordingly.
 *
 * \param tgafile The FILE to read the header from.
 * \param data    The image structure to set the width, height, and bpp in.
 * \param invert  This will be set to true if the origin is in the top 
 *                right rather than bottom left.
 * \param verbose If true, header information is printed to stdout.
 * \return true on success, false otherwise.
 */
static int load_header(FILE *tgafile, Image *data, int *invert, int verbose)
{
    int8_t headdata[18]; // Some space to load the header into

    // We need to get 18 bytes from the file, fall over if we can't
    if(fread(headdata, 1, 18, tgafile) != 18) {
        fprintf(stderr, "Unable to read 18 bytes of header data.\nError was: %s\n", strerror(errno));
        return 0;
    }

    // Before we do anything else, check that the image is the required type
    if(headdata[2] != 2) {
        fprintf(stderr, "Unsupported TGA format (%d).\nThis program only supports uncompressed RGB(A) images.\n", headdata[2]);
        return 0;
    }

    // Only 16, 24, or 32 bit targa files are supported
    if(headdata[16] != 16 && headdata[16] != 24 && headdata[16] != 32) {
        fprintf(stderr, "Unsupported pixel size (%d).\nThis program only supports 16, 24, or 32 bit TGAs.\n", headdata[16]);
        return 0;
    }

    // Only allow 0 or 1 in the colourmaptype field
    if(headdata[1] != 0 && headdata[1] != 1) {
        fprintf(stderr, "Unsupported colour map type (%d).\nThis program only supports colourmap types 0 and 1.\n",headdata[1]);
        return 0;
    }

    // get here and the tga should be readable. Pull out the image metadata
    data -> originx = (headdata[9]  << 8) | headdata[8];
    data -> originy = (headdata[11] << 8) | headdata[10];
    data -> width   = (headdata[13] << 8) | headdata[12];
    data -> height  = (headdata[15] << 8) | headdata[14];
    data -> bpp     = headdata[16];

    *invert = headdata[17] & 0x20; // if bit 5 is set, the origin is top left

    if(verbose) { 
        printf("Header read successfully.\n Width: %d\nHeight: %d\n   BPP: %d\n", data -> width, data -> height, data -> bpp);
    }

    // Skip image descriptor and colourmap if necessary
    int skip = headdata[0];
    skip += (headdata[1] * ((headdata[6] << 8) | headdata[5]));
    if(skip) {
        if(verbose) {
            printf("Skipping %d bytes for colormap and ID field.\n", skip);
        }
        fseek(tgafile, skip, SEEK_CUR);
    }

    // By this point, the file pointer is in the right place, and the 
    // image data is filled in. Return true...
    return 1;
}


/** Read the contents of the data area of a TGA file into memory. This assumes that
 *  the specified image structure contains a data array large enough to store the 
 *  contents of the image.
 *
 * \param tgafile The file to load the image data from.
 * \param data    The image structure to store the data in.
 * \param invert  If true, the data will be flipped vertically after loading.
 * \param verbose If true, some more information is printed to stdout.
 * \return true on success, false if there is a read error, or the image data available
 *         does not match the image dimensions.
 */
static int load_data(FILE *tgafile, Image *data, int invert, int verbose)
{
    int readcount = 0;

    // 32 bit is le easy - just read straight into memory
    if(data -> bpp == 32) {
        if(verbose) {
            printf("Loading 32 bit image data into memory...");
        }
        readcount = fread(data -> data, sizeof(Pixel), data -> width * data -> height, tgafile);

    // 24 bit needs a bit more work  
    } else if(data -> bpp == 24) {
        if(verbose) {
            printf("Loading 24 bit image data into memory... ");
        }

        while(readcount < (data -> width * data -> height)) {
            int rsize = fread(&data -> data[readcount], 1, 3, tgafile);
            if(rsize != 3) {
                fprintf(stderr, "Truncated read - expected 3 bytes, got %d\n", rsize);
                return 0;
            }
            ++readcount;
        }

    // 16 bit involves converting packed data, yey
    } else if(data -> bpp == 16) {
        if(verbose) {
            printf("Loading 16 bit image data into memory... ");
        }

        uint8_t pixels[2];

        while(readcount < (data -> width * data -> height)) {
            // FIXME: Oh Gods the endian horrors!
            int rsize = fread(pixels, 2, 1, tgafile);
            if(rsize != 2) {
                fprintf(stderr, "Truncated read - expected 2 bytes, got %d\n", rsize);
                return 0;
            }

            // decode into rgb fields. Bit twiddling ahoy.
            // FIXME: probable endian horrors herein, too!
            data -> data[readcount].r = (pixels[1] & 0x7c) << 1;
            data -> data[readcount].g = ((pixels[1] & 0x03) << 6) | ((pixels[0] & 0xe0) >> 2);
            data -> data[readcount].b = (pixels[0] & 0x1f) << 3;
            data -> data[readcount].a = (pixels[1] & 0x80);
            ++readcount;
        }

    // Any other pixel width is verboten
    } else {
        fprintf(stderr, "Unsupported bpp (%d) in load_data. How the hell did this get past header checks?\n", data -> bpp);
        return 0;
    }

    // Bomb if we didn't get enough pixels
    if(readcount != (data -> width * data -> height)) {
        fprintf(stderr, "Image data truncated. Only have %d of expected %d pixels.\n", readcount, data -> width * data -> height);
        return 0;
    }

    // Get here and everything is okay
    if(verbose) {
        printf("read complete.\nRead %d (%d x %d) pixels into memory.\n", readcount, data -> width, data -> height);
    }

    // Do we need to invert on the Y?
    if(invert) {
        if(verbose) printf("Image needs to be flipped vertically. Flipping... ");

        flip_vertically(data);

        if(verbose) printf("done.\n");
    }

    return 1;
}


Image *loadTGA(char *filename, int verbose)
{
    Image *newimage;
    FILE  *imagefile;
    int    needinvert;

    // First, try to open the file
    if(imagefile = fopen(filename, "r")) {
        
        // Attempt to allocate memory for the image structure
        if(newimage = calloc(1, sizeof(Image))) {

            // Load the header
            if(load_header(imagefile, newimage, &needinvert, verbose)) {

                // Allocate space for the image data
                if(newimage -> data = calloc(newimage -> width * newimage -> height, sizeof(Pixel))) {

                    // Try to read in the image data
                    if(load_data(imagefile, newimage, needinvert, verbose)) {
                        // Everything is loaded - close the file and return the image
                        fclose(imagefile);
                        return newimage;
                    } 
                    
                    // loading failed, so drop the image data memory
                    free(newimage -> data);
                } else { // if(newimage -> data = calloc(newimage -> width * newimage -> height, sizeof(Pixel))) {
                    fprintf(stderr, "Unable to allocate %d bytes for image data.",newimage -> width * newimage -> height * sizeof(Pixel));
                }
            } // if(load_header(imagefile, newimage)) {

            // On the way out, so free up the memory
            free(newimage);
        } else { // if(newimage = calloc(1, sizeof(Image))) {
            fprintf(stderr, "Unable to allocate %d bytes for a new image structure", sizeof(Image));
        }
        fclose(imagefile);
    } else { // if(imagefile = fopen(filename, "r")) {
        fprintf(stderr, "Unable to open tga file '%s': %s\n", filename, strerror(errno));
    }

    // We fail at loading...
    return NULL;
}
    

int saveTGA(char *filename, Image *data, int verbose)
{
    int result = 1;
    int pixel;

    // empty header storage 
    uint8_t header[18] = { 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

    // work out the depth (we don't do 16 bit - that becomes 32)
    if(data -> bpp == 16) data -> bpp = 32;

    // fill in appropriate fields
    header[ 8] = (data -> originx & 0x00FF);       // x origin
    header[ 9] = (data -> originx & 0xFF00) >> 8;
    header[10] = (data -> originy & 0x00FF);       // y origin
    header[11] = (data -> originy & 0xFF00) >> 8; 
    header[12] = (data -> width   & 0x00FF);       // width
    header[13] = (data -> width   & 0xFF00) >> 8;
    header[14] = (data -> height  & 0x00FF);       // height
    header[15] = (data -> height  & 0xFF00) >> 8;
    header[16] = data -> bpp;                      // depth

    // For 32 bit images, bits 0 to 3 of the ID byte must be set
    if(data -> bpp == 32) header[17] = 0x08;

    if(verbose) printf("Saving TGA file '%s'... ", filename);

    // Now we can start file work - open the file
    FILE *outfile = fopen(filename, "w");
    if(outfile) {

        // Write the header
        if(fwrite(header, 1, 18, outfile) == 18) {

            // Now we need to write the data out
            for(pixel = 0; pixel < (data -> width * data -> height); ++pixel) {
                if(fwrite(&data -> data[pixel], data -> bpp / 8, 1, outfile) == 0) {
                    fprintf(stderr, "Unable to write image data: %s\n", strerror(errno));
                    result = 0;
                    break;
                }
            }
            
        } else { // if(fwrite(header, 1, 18, outfile) == 18) {
            fprintf(stderr, "Unable to write the tga header: %s\n", strerror(errno));
            result = 0;
        }   

        fclose(outfile);
    } else { // if(outfile) {
        fprintf(stderr, "Unable to open '%s' for writing: %s\n", filename, strerror(errno));
        result = 0;
    }
        
    if(verbose && result) printf("done.\n");

    return result;
}


void free_image(Image *data)
{
    // First, nuke the data inside the data if it is present
    if(data -> data) free(data -> data);
    
    // Now nuke the image
    free(data);
}
