/** @file targa.h
 *  Header file for Targa loading and saving code.
 *  This file contains structures and prototypes related to loading and saving Truevision 
 *  Targa files.
 *
 *  \author Chris <chris@starforge.co.uk>
 *  \copy 2008, Chris Page &lt;chris@starforge.co.uk&gt;
 *  \version 1.0
 *  \date 2008-10-30
 */
/*  
 *  Copyright (C) 2008 Chris Page <chris@starforge.co.uk>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be entertaining,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.  A copy of the
 *  General Public License is included in the file COPYING.
 */

#ifndef _TARGA_H
#define _TARGA_H 1

#include <stdint.h>


/** A structure representing a single pixel in the image, containing data in
 *  B8G8R8A8 format (the channel ordering used in TGA files).
 */
typedef struct {
    uint8_t b;   //!< Pixel blue channel.
    uint8_t g;   //!< Pixel green channel.
    uint8_t r;   //!< Pixel red channel.
    uint8_t a;   //!< Pixel alpha channel.
} Pixel;


/** Structure containing image data. This structure is created by a loadTGA()
 *  call, and should be passed to a saveTGA() call.
 */
typedef struct {
    int    originx;  //!< Image x origin
    int    originy;  //!< Image y origin
    int    width;    //!< The width of the image in pixels.
    int    height;   //!< The height of the image in pixels.
    int    bpp;      //!< Number of bits per pixel in the image (should be 24 or 32).
    Pixel *data;     //!< A pointer to width * height pixels of data for the image.
} Image;


/** Load a TGA file into memory. This function will only support loading of 16, 24
 *  or 32 bit type 2 (uncompressed RGB) targa files. If the specified file is not 
 *  in the a supported format, or a problem occurs, this will return NULL and print 
 *  an error.
 *
 * \param filename The name of the targa file to load into memory.
 * \param verbose  Set to true to output additional information.
 * \return An Image structure containing the data loaded from the file, or NULL
 *         if the file is in an unsupported format, or an error occured.
 */
Image *loadTGA(char *filename, int verbose);


/** Save an image in memory to a file as uncompressed 24 bit RGB or 32 bit RGBA
 *  targa file.
 *
 * \param filename The name of the targa file to write to. If the file exists, it will be
 *                 overwritten!
 * \param data     The image data to write to the file.
 * \param verbose  Set to true to output additional information.
 * \return true if the image was saved successfully, false if an error occured.
 */ 
int saveTGA(char *filename, Image *data, int verbose);


/** Release the memory occupied by an Image. After this is called, the memory pointed
 *  to by data is no longer valid, and pointers to it should be discarded.
 *
 * \param data A pointer to the image to free.
 */
void free_image(Image *data);

#endif
